//RandomAgar v1.1
//(c) 2002 Gabriel Nivasch

template <class Symm>class Field
{
	public:
		//Maximum field size:
		const static int XMAX=28, YMAX=35;
		//Fixed period to check for when stepping:
		const static int fixed_p=2;
		//CA rule:
		const static int CArule[2][9]; //defined below
		
		//Default constructor:
		Field():celllist(NULL), currg(0)
		{}
		
		//Copy constructor (we have to override the default behavior).
		//Call our operator=() function:
		Field(const Field &field2):celllist(NULL), currg(0)
		{ *this=field2; }
		
		//Generate a random symmetry and link the cells according
		//to it:
		Symm randomizeSymm()
		{
			symm=Symm::getRandomSymm(XMAX, YMAX);
			buildnet();
			return symm;
		}
		
		//Initialize the field to a random initial pattern,
		//and clear all the other generations:
		void randomizeField()
		{
			//Cells should already be linked:
			assert(celllist);
			
			currg=0;
			pop=0;
			for (cl *cp=celllist; cp!=NULL; cp=cp->next)
			{
				bool alive=(RM.rand()%3==0); //Prob. 1/3
				cp->g[0]=alive;
				pop+=alive;
				for (int gen=1; gen<fixed_p; gen++)
					cp->g[gen]=0;
			}
		}
		
		//Step the pattern one generation. Return value:
		//1 if the pattern died.
		//2 if the pattern became stable.
		//3 if the fixed period was detected.
		//0 otherwise.
		int step()
		{
			//Cells should already be linked:
			assert(celllist);
			
			int nextg=(currg+1)%fixed_p;
			bool stable=true, periodic=true;
			
			pop=0;
			for (cl *cp=celllist; cp!=NULL; cp=cp->next)
			{
				int center=cp->g[currg],
					nbrs=cp->n->g[currg] +
				         cp->ne->g[currg] +
				         cp->e->g[currg] +
				         cp->se->g[currg] +
				         cp->s->g[currg] +
				         cp->sw->g[currg] +
				         cp->w->g[currg] +
				         cp->nw->g[currg];
				assert(center==0 || center==1);
				assert(nbrs>=0 && nbrs<=8);
				
				int nextval=CArule[center][nbrs];
				
				if (nextval!=cp->g[currg]) stable=false;
				if (nextval!=cp->g[nextg]) periodic=false;
					
				cp->g[nextg]=nextval;
				pop+=nextval;
			} //for cl
			currg=nextg;
			
			if (pop==0) return 1;
			if (stable) return 2;
			if (periodic) return 3;
			return 0;
		} //step()
		
		//Get current population.
		int getPop()
		{ return pop; }
		
		//Return true if the pattern is completely made of orthogonal or
		//diagonal lines (which is boring).
		bool isStringy()
		{
			bool horiz=true, vert=true, diag1=true, diag2=true;
			for (cl *cp=celllist; cp!=NULL; cp=cp->next)
			{
				if (cp->g[currg] != cp->e ->g[currg] ||
					cp->g[currg] != cp->w ->g[currg]) horiz=false;
				if (cp->g[currg] != cp->n ->g[currg] ||
					cp->g[currg] != cp->s ->g[currg]) vert=false;
				if (cp->g[currg] != cp->ne->g[currg] ||
					cp->g[currg] != cp->sw->g[currg]) diag1=false;
				if (cp->g[currg] != cp->se->g[currg] ||
					cp->g[currg] != cp->nw->g[currg]) diag2=false;
			}
			return horiz || vert || diag1 || diag2;
		}
		
		//Duplicate another field's state (all generations):
		Field<Symm> &operator =(const Field &field2)
		{
			assert(field2.celllist);
			
			//Make sure we have identical symmetries:
			if(celllist==NULL || symm!=field2.symm)
			{
				symm=field2.symm;
				buildnet();
			}
			
			currg=field2.currg;
			pop=field2.pop;
			
			for (cl *cp=celllist, *cp2=field2.celllist; cp!=NULL;
				cp=cp->next, cp2=cp2->next)
				for (int gen=0; gen<fixed_p; gen++)
					cp->g[gen]=cp2->g[gen];
			return *this;
		}
		
		//Compare two fields (just their current generations):
		bool operator ==(Field &field2)
		{
			assert(celllist);
			assert(field2.celllist);
			if (symm!=field2.symm) return false;
			
			//First compare populations:
			if (pop!=field2.pop) return false;
			
			for (cl *cp=celllist, *cp2=field2.celllist; cp!=NULL;
				cp=cp->next, cp2=cp2->next)
				if (cp->g[currg]!=cp2->g[field2.currg])
					return false;
			
			return true;
		}
		
		bool operator !=(Field &field2)
		{ return !(*this==field2); }
		
		//Get a hash value for the current generation.
		//The population count is used as the most significant bits, so
		//that patterns with less population have smaller hash values.
		unsigned long long getHash()
		{
			assert(celllist);
			//Hash symmetry:
			unsigned long long h=symm.getHash();
			//Hash cell states:
			for (cl *cp=celllist; cp!=NULL; cp=cp->next)
				h = 31*h + cp->g[currg];
			//Add population to hash:
			const int popBits=12; //Max population is 4095
			assert(pop < (1<<popBits));
			h = (h & ((unsigned long long)(-1)>>popBits)) |
				((unsigned long long)pop<<(64-popBits));
			
			return h;
		}
		
		bool operator <(Field &field2)
		{ return getHash()<field2.getHash(); }
		
		bool operator >(Field &field2)
		{ return field2<*this; }
		
		//Print a section of the agar to the specified file:
		void printAgar(FILE *f)
		{
			int xmax=symm.getTileX(), ymax=symm.getTileY();
			fprintf(f, "#LLAB: Torus=%dx%d TopLeft=1,1\n"
				"x = %d, y = %d", ymax, xmax, xmax, ymax);
			
			char rule[30];
			getRuleString(rule);
			if (rule[0]!='\0')
				fprintf(f, ", rule = %s", rule);
			fprintf(f, "\n");
			
			RLEdoer rle;
			char s[20];
			for (int y=0; y<ymax; y++)
			{
				for (int x=0; x<xmax; x++)
				{
					int xc, yc;
					symm.getcell(x, y, &xc, &yc);
					rle.doRLE(cell[xc][yc].g[currg] ? RLEdoer::alive :
						RLEdoer::dead, s);
					fprintf(f, s);
				}
				rle.doRLE(RLEdoer::newline, s);
				fprintf(f, s);
			}
			rle.endRLE(s);
			fprintf(f, "%s\n", s);
		}
		
		//Print the specification and contents of the field in a compact
		//form (just the current generation).
		void printCompactField(FILE *f)
		{
			char c[200];
			symm.getparams(c);
			fprintf(f, "#C Symmetry %d. Params: %s.", symm.getID(), c);
			
			//Count the number of valid cells:
			cl *cp;
			int ncells=0;
			for (cp=celllist; cp!=NULL; cp=cp->next)
				ncells++;
			fprintf(f, " %d cells:", ncells);
			
			//Print the cell values in hexadecimal:
			int nbits=0, chr=0;
			for (cp=celllist; cp!=NULL; cp=cp->next)
			{
				if (nbits % (70*4) == 0)
					fprintf(f, "\n#C ");
				chr = 2*chr + cp->g[currg];
				nbits++;
				if (nbits % 4 == 0)
				{
					fprintf(f, "%X", chr);
					chr=0;
				}
			}
			if (nbits % 4)
				fprintf(f, "%X", chr);
			fprintf(f, "\n");
		} //printCompactField()
		
		//We assume the file has been read until just before the symmetry
		//parameters.
		void readCompactField(FILE *f)
		{
			//Read symmetry parameters until the period:
			int n=0;
			char c, params[30];
			while ((c=fgetc(f))!='.')
			{ assert(!feof(f)); params[n++]=c; }
			params[n]='\0';
			
			//Adopt symmetry:
			symm=Symm(XMAX, YMAX, params);
			buildnet();
			
			//Read number of cells:
			int ncells;
			fscanf(f, "%d", &ncells);
			
			//Read until newline:
			while(fgetc(f)!='\n') { assert(!feof(f)); }
			//Skip 3 chars ("#C "):
			fgetc(f); fgetc(f); fgetc(f);
			
			currg=pop=0;
			//Read cell values in hexadecimal and set the field:
			int nbits=0, val;
			char chr[2]={'\0', '\0'};
			for (cl *cp=celllist; cp!=NULL; cp=cp->next)
			{
				assert(nbits<ncells);
				if (nbits%4==0)
				{
					//Read another char:
					assert(!feof(f));
					chr[0]=fgetc(f);
					if (chr[0]=='\n') //Skip 3 more chars ("#C ")
					{ fgetc(f); fgetc(f); fgetc(f); chr[0]=fgetc(f); }
					//Convert char to integer:
					sscanf(chr, "%x", &val);
					assert(val>=0 && val<16);
					//Adapt to quirk in last character:
					if (ncells-nbits<4)
						val <<= 4-ncells+nbits;
				}
				
				//Get cell value:
				bool alive = (val&8) != 0;
				val<<=1;
				
				//Set cell:
				cp->g[0]=alive;
				pop+=alive;
				for (int gen=1; gen<fixed_p; gen++)
					cp->g[gen]=0;
				
				nbits++;
			} //for cl
			
			assert(nbits==ncells);
		} //readCompactField()
		
		//Debugging function:
		void printValidField(FILE *f)
		{
			int maxvalidx=-1, maxvalidy=-1;
			
			for (int y=0; y<YMAX; y++)
			{
				for (int x=0; x<XMAX; x++)
					if (cell[x][y].valid)
					{
						if (x>maxvalidx) maxvalidx=x;
						if (y>maxvalidy) maxvalidy=y;
						fprintf(f, "#");
					}
					else
						fprintf(f, ",");
				fprintf(f, "\n");
			}
			fprintf(f, "Open bound: (%d, %d)\n", maxvalidx+1, maxvalidy+1);
		}
				
	
	private:
		//Cell structure:
		struct cl
		{
			int valid, g[fixed_p];
			cl *n, *ne, *e, *se, *s, *sw, *w, *nw, *next;
		};
		//The cells:
		cl cell[XMAX][YMAX];
		//Linked list of valid cells:
		cl *celllist;
		//Current generation (0 to fixed_p-1):
		int currg;
		//Population in current generation:
		int pop;
		//The symmetry specification:
		Symm symm;
		
		//Link the cells together according to the symmetry.
		//Also build the linked list of valid cells:
		void buildnet()
		{
			assert(symm.getxmax()<=XMAX && symm.getymax()<=YMAX);
			
			cl **listend=&celllist;
			
			for (int i=0; i<XMAX; i++)
				for (int j=0; j<YMAX; j++)
				{
					int i2, j2;
					
					//Check if cell (i, j) is valid:
					symm.getcell(i, j, &i2, &j2);
					if (i!=i2 || j!=j2)
					{
						//(cell is invalid)
						cell[i][j].valid=0;
						continue;
					}
					//Cell is valid, so add it to the linked list
					//through the "next" pointer:
					cell[i][j].valid=1;
					*listend=&cell[i][j];
					listend=&cell[i][j].next;
					
					//Link cell (i, j) to its neighbors:
					//N:
					symm.getcell(i, j-1, &i2, &j2);
					cell[i][j].n=&cell[i2][j2];
					//NE:
					symm.getcell(i+1, j-1, &i2, &j2);
					cell[i][j].ne=&cell[i2][j2];
					//E:
					symm.getcell(i+1, j, &i2, &j2);
					cell[i][j].e=&cell[i2][j2];
					//SE:
					symm.getcell(i+1, j+1, &i2, &j2);
					cell[i][j].se=&cell[i2][j2];
					//S:
					symm.getcell(i, j+1, &i2, &j2);
					cell[i][j].s=&cell[i2][j2];
					//SW:
					symm.getcell(i-1, j+1, &i2, &j2);
					cell[i][j].sw=&cell[i2][j2];
					//W:
					symm.getcell(i-1, j, &i2, &j2);
					cell[i][j].w=&cell[i2][j2];
					//NW:
					symm.getcell(i-1, j-1, &i2, &j2);
					cell[i][j].nw=&cell[i2][j2];
				} //for for
			
			*listend=NULL;
		} //buildnet()
		
		void getRuleString(char *r)
		{
			int n=0, i;
			r[n++]='B';
			for (i=0; i<=8; i++)
				if (CArule[0][i])
					r[n++]='0'+i;
			r[n++]='/';
			r[n++]='S';
			for (i=0; i<=8; i++)
				if (CArule[1][i])
					r[n++]='0'+i;
			r[n]='\0';
			
			//Life is default:
			if (strcmp(r, "B3/S23")==0)
				r[0]='\0';
		}
}; //Field<Symm>

template<class Symm> const int Field<Symm>::CArule[2][9]=
	{{0, 0, 0, 1, 0, 0, 0, 0, 0},  //Birth
	 {0, 0, 1, 1, 0, 0, 0, 0, 0}}; //Survival