//RandomAgar v1.1
//(c) 2002 Gabriel Nivasch
#include "symm.h"
#include "RanMar.h"
#include "field.h"
#include "RLEdoer.h"

//The heavy-duty random number generator:
RanMar RM;

const int
	maxsymms=0,
	fieldspersymm=20,
	printEvery=10,
	minperiod=0;
const bool
	filterStringy=true,
	filterConstPop=true;
const char hashFile[]="index.txt";

void readHashValues(std::set<unsigned long long> *hashes)
{
	unsigned long long hsh;
	FILE *f=fopen(hashFile, "r");
	if (f!=NULL)
	{
		while(1)
		{
			fscanf(f, "%llx", &hsh);
			if (feof(f))
				break;
			hashes->insert(hsh);
		}
		fclose(f);
	}
}

void getUniqueFileName(int per, char *s)
{
	for (int n=1; n<1000; n++)
	{
		if (per==0) sprintf(s, "p??? %03d", n);
		else sprintf(s, "p%07d %03d", per, n);
		FILE *f=fopen(s, "r");
		if (f==NULL) //No such file. Bingo.
			return;
		fclose(f);
	}
	assert(0); //No suitable filename found
}

template<class Symm> void writeAgarToFile(Field<Symm> *fInitial,
	Field<Symm> *f, int per, int gen, int symmn, int fieldn,
	unsigned long long hsh)
{
	char name[40];
	getUniqueFileName(per, name);
	
	FILE *agarf=fopen(name, "w");
	assert(agarf);
	
	fprintf(agarf, "#C symmn = %d. fieldn = %d. Initial pattern:\n",
		symmn, fieldn);
	fInitial->printCompactField(agarf);
	
	if (per==0) //No period detected
		fprintf(agarf, "#C\n#C p??? gen %d:\n", gen);
	else
		fprintf(agarf, "#C\n#C p%d gen %d (hash = %016llx):\n", per, gen, hsh);
	f->printAgar(agarf);
	
	fclose(agarf);
	setCreatorCode(name);
	
	printf("%s\n", name);
} //writeAgarToFile<Symm>()

template<class Symm> void useField()
{
	//Read existing hash values:
	std::set<unsigned long long> hashes;
	readHashValues(&hashes);
	
	FILE *file=fopen(hashFile, "a");
	assert(file);
	
	int symmn, fieldn, gen, i;
	int lastPop, samePopN;
	Field<Symm> f, fInitial;
	//The kept hashes:
	const int minSize=60;
	unsigned long long hshMinima[minSize];
	int genMinima[minSize], minN;
	
	//Main loop:
	for (symmn=0; symmn<maxsymms || maxsymms==0; symmn++)
	{
		f.randomizeSymm();
		if (symmn%printEvery==0)
			printf("symmn=%d...\n", symmn);
		for (fieldn=0; fieldn<fieldspersymm; fieldn++)
		{
			f.randomizeField();
			fInitial=f;
			minN=0;
			lastPop=0;
			for (gen=1; true; gen++)
			{
				//Step the pattern:
				int r=f.step();
				if (r)
					break;
				
				//Keep track of population changes:
				int pop=f.getPop();
				if (pop!=lastPop)
					samePopN=0;
				samePopN++;
				lastPop=pop;
				
				//Save the current hash in its proper place in the array:
				bool periodic=false;
				unsigned long long hsh=f.getHash();
				for (i=0; i<minN && hsh>=hshMinima[i]; i++)
					if (hsh==hshMinima[i])
					{ periodic=true; break; }
				if (!periodic)
				{
					assert(i<minSize);
					hshMinima[i]=hsh;
					genMinima[i]=gen;
					minN=i+1;
					continue;
				}
				
				//Pattern is periodic. Calculate the period:
				int per = gen - genMinima[i];
				
				//Filter out boring patterns:
				if (filterStringy && f.isStringy())
					break;
				if (filterConstPop && per>=10 && samePopN>=per)
					break;
				if (per<minperiod)
					break;
				
				//See if we already know this pattern based on the hash value:
				if (hashes.find(hsh) != hashes.end())
					break; //Already known
				
				//Write the agar into a file:
				writeAgarToFile(&fInitial, &f, per, gen, symmn, fieldn, hsh);
				
				//Remember the hash value:
				hashes.insert(hsh);
				fprintf(file, "%016llx\n", hsh);
				fflush(file);
				
				break; //the "for gen" loop
			} //for gen
		} // for fieldn
	} //for symmn
	
	fclose(file);
} //useField<Symm>()

int main()
{
	//Initialize random number generator:
	int s1, s2;
	RM.genSeeds(&s1, &s2);
	RM.srand(s1, s2);
	
	printf("XMAX==%d, YMAX==%d. Random seeds: %d, %d\n\n", Field<TorusSymm>::XMAX,
		Field<TorusSymm>::YMAX, s1, s2);
	printf("Please choose the symmetry type:\n"
		"(1) torus\n"
		"(2) closed pillowcase\n"
		"(3) orthogonal cylinder                 (-3) diagonal cylinder\n"
		"(4) orthogonal Klein bottle             (-4) diagonal Klein bottle\n"
		"(5) orthogonal rectangle                (-5) diagonal rectangle\n"
		"(6) orthogonal open pillowcase          (-6) diagonal open pillowcase\n"
		"(7) orthogonal nonorientable football   (-7) diagonal nonorientable football\n"
		"(8) orthogonal Mobius strip             (-8) diagonal Mobius strip\n"
		"(9) or. turnover slit along hypotenuse  (-9) di. turnover slit along hypotenuse\n"
		"(10) turnover\n"
		"(11) orthogonal triangle                (-11) diagonal triangle\n"
		"(12) orthogonal turnover slit along leg (-12) diagonal turnover slit along leg\n"
		"\n");
	int sn;
	scanf("%d", &sn);
	
	//Branch according to symmetry:
	try
	{
		switch(sn)
		{
			case   1: useField<TorusSymm>(); break;
			case   2: useField<ClosedPillowcaseSymm>(); break;
			case   3: useField<OrthoCylinderSymm>(); break;
			case  -3: useField<DiagCylinderSymm>(); break;
			case   4: useField<OrthoKleinSymm>(); break;
			case  -4: useField<DiagKleinSymm>(); break;
			case   5: useField<OrthoRectangleSymm>(); break;
			case  -5: useField<DiagRectangleSymm>(); break;
			case   6: useField<OrthoOpenPillowcaseSymm>(); break;
			case  -6: useField<DiagOpenPillowcaseSymm>(); break;
			case   7: useField<OrthoFootballSymm>(); break;
			case  -7: useField<DiagFootballSymm>(); break;
			case   8: useField<OrthoMobiusSymm>(); break;
			case  -8: useField<DiagMobiusSymm>(); break;
			case   9: useField<OrthoTurnoverHypoSymm>(); break;
			case  -9: useField<DiagTurnoverHypoSymm>(); break;
			case  10: useField<TurnoverSymm>(); break;
			case  11: useField<OrthoTriangleSymm>(); break;
			case -11: useField<DiagTriangleSymm>(); break;
			case  12: useField<OrthoTurnoverLegSymm>(); break;
			case -12: useField<DiagTurnoverLegSymm>(); break;
			default: printf("Invalid input.\n"); break;
		}
	}
	catch(std::bad_alloc)
	{ assert(0); } //Probably the set ran out of space
	catch(...)
	{ assert(0); }
	
	printf("Done.\n");
}