//RandomAgar v1.1
//(c) 2002 Gabriel Nivasch
#include "symm.h"
#include "RanMar.h"

extern RanMar RM;

//Mac-specific code - make given file open with LifeLab:
void setCreatorCode(const char *name)
{
	Str255 plname;
	CopyCStringToPascal(name, plname);
	FSSpec spc;
	int err=FSMakeFSSpec(0, 0, plname, &spc);
	assert(err==0);
	FInfo theinfo;
	err=FSpGetFInfo(&spc, &theinfo);
	assert(err==0);
	theinfo.fdCreator='LLAB';
	err=FSpSetFInfo(&spc, &theinfo);
	assert(err==0);
}

int min(int a, int b)
{ return a<b?a:b; }

int gcd(int a, int b)
{
	int c=a%b;
	while (c!=0)
	{ a=b; b=c; c=a%b; }
	return b;
}

void pickSmaller(int *x1, int *y1, int x2, int y2)
{
	if (x2<*x1 || (x2==*x1 && y2<*y1))
	{ *x1=x2; *y1=y2; }
}

//Given vectors of repetition (x1, y1) and (x2, y2), map the input
//point to a point within the "basic area", such that all points that are
//equivalent are mapped to the same output point.
//The output range is nonnegative and as compact as possible.
//The caller has to make sure that the vectors of repetition are in the
//form 0<=x1<x2, 0<=y2<y1.
//(The vectors of repetition can always be converted to this form -
//sometimes in more than one way.)
void moveToBasicArea(int x1, int y1, int x2, int y2,
	int inx, int iny, int *outx, int *outy)
{
	assert(0<=x1 && x1<x2 && 0<=y2 && y2<y1);
	
	//Move the input point to the basic rhomboid
	//(we use floor((double)) because it works for both positive
	//and negative integers):
	int k1=floor((double)(inx*y2-iny*x2)/(x1*y2-x2*y1)),
		k2=floor((double)(inx*y1-iny*x1)/(x2*y1-x1*y2));
	inx -= k1*x1+k2*x2;
	iny -= k1*y1+k2*y2;
	
	//Move the point closer to the origin if possible:
	if (inx>=x1 && iny>=y1)
	{ inx-=x1; iny-=y1; }
	if (inx>=x2 && iny>=y2)
	{ inx-=x2; iny-=y2; }
	
	*outx=inx;
	*outy=iny;
	
	assert(*outx>=0 && *outx<x2 && *outy>=0 && *outy<y1);
}

//From here on is the implementation of the symmetries.
//
//In general, there are 17 symmetries that a periodic planar pattern
//can have. They are known as the 17 "wallpaper groups".
//
//I followed the order in Xah Lee's web page, "The 17 Wallpaper
//Groups":
//http://www.xahlee.org/Wallpaper_dir/c5_17WallpaperGroups.html
//All the references here are to it. (If the page goes out for any
//reason, I have a local copy.)
//
//Symmetries 13 through 17 have 120-degree rotations, so they don't
//apply to Life. Of the remaining ones, symmetries 1, 2, and 10 can
//have an arbitrary orientation. The other ones (3, 4, 5, 6, 7, 8, 9,
//11, and 12), can have two orientations: orthogonal or 45-degree
//diagonal.
//
//The symmetries further branch according to "parity": Centers of
//180-degree rotation can lie at cell centers, at cell corners, or
//at horizontal or vertical edges. Centers of 90-degree rotation
//can lie at cell centers or cell corners. Orthogonal axes (either
//of reflection or glide-reflection) can go through cell centers or
//between cells. Diagonal axes of glide-reflection can go through
//cell centers and corners or through horizontal and vertical
//edges. If two mutually perpendicular diagonal axes go through
//cell centers and corners, they can cross either at a cell center
//or at a cell corner.
//
//I had to give names to the 12 symmetries in a consistent way, so
//I took the names that appear under the column "quotient space":
//"torus", "closed pillowcase", etc.
//
//Each symmetry class has, besides xmax and ymax, some parameters
//that specify the symmetry. The most important function of each class
//is getcell(), which maps a given input cell to the
//corresponding cell in the symmetry's basic area, in such a way
//that all input cells that should be equivalent are mapped to the
//same output cell.
//
//The procedure in getcell() is always the same: Call
//moveToBasicArea() for the input point; apply to the input
//point the different "generating" transformations for the symmetry
//(as shown in the table under the column "Symmetries and generators")
//and call moveToBasicArea() for them; and from all the reduced points,
//choose one of them deterministically: the one with the smallest
//coordinates.
//
//To cover all the parities in each case, the
//symmetries have additional parameters called "center", "axis", etc.,
//that can take values 0 or 1.
//
//The function getRandomSymm() generates random values for the
//parameters, such that the resulting basic area lies within the
//specified maximum values. This implies that the parameters must have
//some constraints; we want to make them as loose as possible, in order
//to use as much as possible of the allowed range. Most of the times, I
//found the constraints experimentally. The constraints are always
//indicated in a comment inside the function.
//
//In some cases (for example, Diagonal Cylinder), the constraints were
//chosen in such a way that they cover only half the possibilities
//(in this case, the vector of repetition parallel to the axis of
//reflection is always the large one). In these cases, in order to cover
//all possibilities, there's an additional "sign" parameter that takes
//the value 1 or -1. Based on it, getcell() decides whether to apply
//the "generating" transformations in one orientation or in the other.
//
//getTileX() and getTileY() return the dimensions of a rectangular patch
//that tiles the plane just by orthogonal translation.

//---------- (1) Torus ------------------------------------------//
TorusSymm::TorusSymm()
{ xmax=ymax=-1; }

TorusSymm::TorusSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d %d", &x1, &y1, &x2, &y2); }

TorusSymm TorusSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints are:
	//0<=x1<x2<=xmax; 0<=y2<y1<=ymax.
	int x1, x2, y1, y2, temp;
	
	x1=RM.rand()%(xmax+1);
	do
	{ x2=RM.rand()%(xmax+1); }
	while (x1==x2);
	if (x1>x2)
	{ temp=x1; x1=x2; x2=temp; }
	
	y1=RM.rand()%(ymax+1);
	do
	{ y2=RM.rand()%(ymax+1); }
	while (y1==y2);
	if (y1<y2)
	{ temp=y1; y1=y2; y2=temp; }
	
	TorusSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.x1=x1; s.y1=y1;
	s.x2=x2; s.y2=y2;
	return s;
}
		
void TorusSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (x1, y1), (x2, y2).
	moveToBasicArea(x1, y1, x2, y2, inx, iny, outx, outy);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void TorusSymm::getparams(char *s)
{ sprintf(s, "%d %d %d %d", x1, y1, x2, y2); }

unsigned long long TorusSymm::getHash()
{ return (unsigned long long)getID() + 10*x1 + 100*y1 + 1000*x2 + 10000*y2; }

int TorusSymm::getTileX()
{ return (x2*y1 - x1*y2)/gcd(y1, y2); }

int TorusSymm::getTileY()
{ return (x2*y1 - x1*y2)/gcd(x1, x2); }

bool TorusSymm::operator ==(const TorusSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& x1==s2.x1 && y1==s2.y1
		&& x2==s2.x2 && y2==s2.y2;
}

//---------- (2) Closed pillowcase ------------------------------//
ClosedPillowcaseSymm::ClosedPillowcaseSymm()
{ xmax=ymax=-1; }

ClosedPillowcaseSymm::ClosedPillowcaseSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d %d %d %d", &x1,&y1,&x2,&y2,&ox,&oy); }

ClosedPillowcaseSymm ClosedPillowcaseSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints are:
	//0<=x1<x2<=xmax; 0<=y2<y1<=ymax.
	int x1, x2, y1, y2, temp, ox, oy;
	
	x1=RM.rand()%(xmax+1);
	do
	{ x2=RM.rand()%(xmax+1); }
	while (x1==x2);
	if (x1>x2)
	{ temp=x1; x1=x2; x2=temp; }
	
	y1=RM.rand()%(ymax+1);
	do
	{ y2=RM.rand()%(ymax+1); }
	while (y1==y2);
	if (y1<y2)
	{ temp=y1; y1=y2; y2=temp; }
	
	ox=RM.rand()%2;
	oy=RM.rand()%2;
	
	ClosedPillowcaseSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.x1=x1; s.y1=y1;
	s.x2=x2; s.y2=y2;
	s.ox=ox; s.oy=oy;
	return s;
}

void ClosedPillowcaseSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (x1, y1), (x2, y2).
	//Center of 180-degree rotation at (ox/2, oy/2).
	moveToBasicArea(x1, y1, x2, y2, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(x1, y1, x2, y2, ox-inx, oy-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void ClosedPillowcaseSymm::getparams(char *s)
{ sprintf(s, "%d %d %d %d %d %d", x1, y1, x2, y2, ox, oy); }

unsigned long long ClosedPillowcaseSymm::getHash()
{ return (unsigned long long)getID()+10*x1+100*y1+1000*x2+10000*y2+100000*ox+1000000*oy; }

int ClosedPillowcaseSymm::getTileX()
{ return (x2*y1 - x1*y2)/gcd(y1, y2); }

int ClosedPillowcaseSymm::getTileY()
{ return (x2*y1 - x1*y2)/gcd(x1, x2); }

bool ClosedPillowcaseSymm::operator ==(const ClosedPillowcaseSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& x1==s2.x1 && y1==s2.y1
		&& x2==s2.x2 && y2==s2.y2
		&& ox==s2.ox && oy==s2.oy;
}

//---------- (3) Orthogonal cylinder ----------------------------//
OrthoCylinderSymm::OrthoCylinderSymm()
{ xmax=ymax=-1; }

OrthoCylinderSymm::OrthoCylinderSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d", &x, &y, &xaxis); }

OrthoCylinderSymm OrthoCylinderSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<x<=2*xmax-2; 0<y<=ymax.
	int x, y, xaxis;
	
	x=1+RM.rand()%(2*xmax-2);
	y=1+RM.rand()%ymax;
	xaxis=RM.rand()%2;
	
	OrthoCylinderSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.x=x; s.y=y; s.xaxis=xaxis;
	return s;
}

void OrthoCylinderSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (x, 0), (0, y).
	//Axis of reflection at x = xaxis/2.
	moveToBasicArea(0, y, x, 0, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(0, y, x, 0, xaxis-inx, iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void OrthoCylinderSymm::getparams(char *s)
{ sprintf(s, "%d %d %d", x, y, xaxis); }

unsigned long long OrthoCylinderSymm::getHash()
{ return (unsigned long long)getID() + 10*x + 100*y + 1000*xaxis; }

int OrthoCylinderSymm::getTileX()
{ return x; }

int OrthoCylinderSymm::getTileY()
{ return y; }

bool OrthoCylinderSymm::operator ==(const OrthoCylinderSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& x==s2.x && y==s2.y && xaxis==s2.xaxis;
}

//---------- (-3) Diagonal cylinder -----------------------------//
DiagCylinderSymm::DiagCylinderSymm()
{ xmax=ymax=-1; }

DiagCylinderSymm::DiagCylinderSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d", &d1, &d2, &dsign); }

DiagCylinderSymm DiagCylinderSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints are:
	//0<d1<=d2<=ymax; d1+d2<=xmax.
	int d1, d2, dsign;
	
	int m=min(xmax/2, ymax);
	d1=1+RM.rand()%m;
	d2=1+RM.rand()%m;
	if (d1>d2)
	{ int temp=d1; d1=d2; d2=temp; }
	dsign=1-2*(RM.rand()%2); //1 or -1
	
	DiagCylinderSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.d1=d1; s.d2=d2; s.dsign=dsign;
	return s;
}

void DiagCylinderSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (d1, -d1), (d2, d2).
	//If dsign==1, axis of reflection at y = x;
	//   else axis of reflection at x + y = 0.
	moveToBasicArea(d2, d2, d1+d2, d2-d1, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(d2, d2, d1+d2, d2-d1, dsign*iny, dsign*inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void DiagCylinderSymm::getparams(char *s)
{ sprintf(s, "%d %d %d", d1, d2, dsign); }

unsigned long long DiagCylinderSymm::getHash()
{ return (unsigned long long)getID() + 10*d1 + 100*d2 + 1000*dsign; }

int DiagCylinderSymm::getTileX()
{ return 2*d1*d2/gcd(d1, d2); }

int DiagCylinderSymm::getTileY()
{ return getTileX(); }

bool DiagCylinderSymm::operator ==(const DiagCylinderSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& d1==s2.d1 && d2==s2.d2 && dsign==s2.dsign;
}

//---------- (4) Orthogonal Klein bottle ------------------------//
OrthoKleinSymm::OrthoKleinSymm()
{ xmax=ymax=-1; }

OrthoKleinSymm::OrthoKleinSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d", &x, &y, &xaxis); }

OrthoKleinSymm OrthoKleinSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<x<=2*xmax-2; 0<y<=ymax/2.
	int x, y, xaxis;
	
	x=1+RM.rand()%(2*xmax-2);
	y=1+RM.rand()%(ymax/2);
	xaxis=RM.rand()%2;
	
	OrthoKleinSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.x=x; s.y=y; s.xaxis=xaxis;
	return s;
}

void OrthoKleinSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (x, 0), (0, 2*y).
	//Axis of glide-reflection at x = xaxis/2.
	moveToBasicArea(0, 2*y, x, 0, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(0, 2*y, x, 0, xaxis-inx, iny+y, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void OrthoKleinSymm::getparams(char *s)
{ sprintf(s, "%d %d %d", x, y, xaxis); }

unsigned long long OrthoKleinSymm::getHash()
{ return (unsigned long long)getID() + 10*x + 100*y + 1000*xaxis; }

int OrthoKleinSymm::getTileX()
{ return x; }

int OrthoKleinSymm::getTileY()
{ return 2*y; }

bool OrthoKleinSymm::operator ==(const OrthoKleinSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& x==s2.x && y==s2.y && xaxis==s2.xaxis;
}

//---------- (-4) Diagonal Klein bottle -------------------------//
DiagKleinSymm::DiagKleinSymm()
{ xmax=ymax=-1; }

DiagKleinSymm::DiagKleinSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d", &d1, &d2, &dsign); }

DiagKleinSymm DiagKleinSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints are:
	//0<d1<=d2<=ymax; d1+d2<=xmax.
	//(It seems they could be relaxed a bit, but I don't know how.)
	int d1, d2, dsign;
	
	int m=min(xmax/2, ymax);
	d1=1+RM.rand()%m;
	d2=1+RM.rand()%m;
	if (d1>d2)
	{ int temp=d1; d1=d2; d2=temp; }
	dsign=1-2*(RM.rand()%2); //1 or -1
	
	DiagKleinSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.d1=d1; s.d2=d2; s.dsign=dsign;
	return s;
}

void DiagKleinSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (d1, -d1), (d2, d2).
	//If dsign==1, axis of glide-reflection at y = x + d2/2,
	//   else, axis of glide-reflection at x + y = -d1/2.
	moveToBasicArea(d2, d2, d1+d2, d2-d1, inx, iny, outx, outy);
	int dn = (dsign==1) ? d2 : d1, outx2, outy2;
	moveToBasicArea(d2, d2, d1+d2, d2-d1, dsign*iny, dsign*inx+dn, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void DiagKleinSymm::getparams(char *s)
{ sprintf(s, "%d %d %d", d1, d2, dsign); }

unsigned long long DiagKleinSymm::getHash()
{ return (unsigned long long)getID() + 10*d1 + 100*d2 + 1000*dsign; }

int DiagKleinSymm::getTileX()
{ return 2*d1*d2/gcd(d1, d2); }

int DiagKleinSymm::getTileY()
{ return getTileX(); }

bool DiagKleinSymm::operator ==(const DiagKleinSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& d1==s2.d1 && d2==s2.d2 && dsign==s2.dsign;
}

//---------- (5) Orthogonal rectangle ---------------------------//
OrthoRectangleSymm::OrthoRectangleSymm()
{ xmax=ymax=-1; }

OrthoRectangleSymm::OrthoRectangleSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d %d", &x, &y, &xaxis, &yaxis); }

OrthoRectangleSymm OrthoRectangleSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints are:
	//0<x<=2*xmax-2; 0<y<=2*ymax-2.
	int x, y, xaxis, yaxis;
	
	x=1+RM.rand()%(2*xmax-2);
	y=1+RM.rand()%(2*ymax-2);
	xaxis=RM.rand()%2;
	yaxis=RM.rand()%2;
	
	OrthoRectangleSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.x=x; s.y=y;
	s.xaxis=xaxis; s.yaxis=yaxis;
	return s;
}

void OrthoRectangleSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (x, 0), (0, y).
	//Axes of reflection at x = xaxis/2, y = yaxis/2.
	moveToBasicArea(0, y, x, 0, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(0, y, x, 0, xaxis-inx, iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, y, x, 0, inx, yaxis-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, y, x, 0, xaxis-inx, yaxis-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void OrthoRectangleSymm::getparams(char *s)
{ sprintf(s, "%d %d %d %d", x, y, xaxis, yaxis); }

unsigned long long OrthoRectangleSymm::getHash()
{ return (unsigned long long)getID() + 10*x + 100*y + 1000*xaxis + 10000*yaxis; }

int OrthoRectangleSymm::getTileX()
{ return x; }

int OrthoRectangleSymm::getTileY()
{ return y; }

bool OrthoRectangleSymm::operator ==(const OrthoRectangleSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax && x==s2.x
		&& y==s2.y && xaxis==s2.xaxis && yaxis==s2.yaxis;
}

//---------- (-5) Diagonal rectangle ----------------------------//
DiagRectangleSymm::DiagRectangleSymm()
{ xmax=ymax=-1; }

DiagRectangleSymm::DiagRectangleSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d", &d1, &d2, &daxis); }

DiagRectangleSymm DiagRectangleSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints are:
	//0<d1<=d2<=ymax; d1+d2<=xmax.
	int d1, d2, daxis;
	
	int m=min(xmax/2, ymax);
	d1=1+RM.rand()%m;
	d2=1+RM.rand()%m;
	if (d1>d2)
	{ int temp=d1; d1=d2; d2=temp; }
	daxis=RM.rand()%2;
	
	DiagRectangleSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.d1=d1; s.d2=d2; s.daxis=daxis;
	return s;
}

void DiagRectangleSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (d1, -d1), (d2, d2).
	//Axes of reflection at y = x, x + y = daxis.
	moveToBasicArea(d2, d2, d1+d2, d2-d1, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(d2, d2, d1+d2, d2-d1, iny, inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(d2, d2, d1+d2, d2-d1, daxis-inx, daxis-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(d2, d2, d1+d2, d2-d1, daxis-iny, daxis-inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void DiagRectangleSymm::getparams(char *s)
{ sprintf(s, "%d %d %d", d1, d2, daxis); }

unsigned long long DiagRectangleSymm::getHash()
{ return (unsigned long long)getID() + 10*d1 + 100*d2 + 1000*daxis; }

int DiagRectangleSymm::getTileX()
{ return 2*d1*d2/gcd(d1, d2); }

int DiagRectangleSymm::getTileY()
{ return getTileX(); }

bool DiagRectangleSymm::operator ==(const DiagRectangleSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& d1==s2.d1 && d2==s2.d2 && daxis==s2.daxis;
}

//---------- (6) Orthogonal open pillowcase ---------------------//
OrthoOpenPillowcaseSymm::OrthoOpenPillowcaseSymm()
{ xmax=ymax=-1; }

OrthoOpenPillowcaseSymm::OrthoOpenPillowcaseSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d %d", &x, &y, &xaxis, &yaxis); }

OrthoOpenPillowcaseSymm OrthoOpenPillowcaseSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<x<=2*xmax-2; 0<y<ymax.
	int x, y, xaxis, yaxis;
	
	x=1+RM.rand()%(2*xmax-2);
	y=1+RM.rand()%(ymax-1);
	xaxis=RM.rand()%2;
	yaxis=RM.rand()%2;
	
	OrthoOpenPillowcaseSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.x=x; s.y=y;
	s.xaxis=xaxis; s.yaxis=yaxis;
	return s;
}

void OrthoOpenPillowcaseSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (x, 0), (0, 2*y).
	//Axis of glide-reflection at x = xaxis/2.
	//Axis of reflection at y = yaxis/2.
	moveToBasicArea(0, 2*y, x, 0, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(0, 2*y, x, 0, xaxis-inx, iny+y, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*y, x, 0, inx, yaxis-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*y, x, 0, xaxis-inx, yaxis-iny-y, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void OrthoOpenPillowcaseSymm::getparams(char *s)
{ sprintf(s, "%d %d %d %d", x, y, xaxis, yaxis); }

unsigned long long OrthoOpenPillowcaseSymm::getHash()
{ return (unsigned long long)getID() + 10*x + 100*y + 1000*xaxis + 10000*yaxis; }

int OrthoOpenPillowcaseSymm::getTileX()
{ return x; }

int OrthoOpenPillowcaseSymm::getTileY()
{ return 2*y; }

bool OrthoOpenPillowcaseSymm::operator ==(const OrthoOpenPillowcaseSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax && x==s2.x
		&& y==s2.y && xaxis==s2.xaxis && yaxis==s2.yaxis;
}

//---------- (-6) Diagonal open pillowcase ----------------------//
DiagOpenPillowcaseSymm::DiagOpenPillowcaseSymm()
{ xmax=ymax=-1; }

DiagOpenPillowcaseSymm::DiagOpenPillowcaseSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d %d", &d1, &d2, &dsign, &daxis); }

DiagOpenPillowcaseSymm DiagOpenPillowcaseSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<d1<=d2<=xmax-2; d2<=ymax.
	int d1, d2, dsign, daxis;
	
	int m=min(xmax-2, ymax);
	d1=1+RM.rand()%m;
	d2=1+RM.rand()%m;
	if (d1>d2)
	{ int temp=d1; d1=d2; d2=temp; }
	dsign=1-2*(RM.rand()%2); //1 or -1
	daxis=RM.rand()%2;
	
	DiagOpenPillowcaseSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.d1=d1; s.d2=d2;
	s.dsign=dsign; s.daxis=daxis;
	return s;
}

void DiagOpenPillowcaseSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (d1, -d1), (d2, d2).
	//If dsign==1, axis of glide-reflection at y = x + d2/2,
	//      axis of reflection at x + y = daxis;
	//   else, axis of glide-reflection at x + y = -d1/2,
	//      axis of reflection at y = x - daxis.
	moveToBasicArea(d2, d2, d1+d2, d2-d1, inx, iny, outx, outy);
	int dn = (dsign==1) ? d2 : d1, outx2, outy2;
	moveToBasicArea(d2, d2, d1+d2, d2-d1, dsign*iny, dsign*inx+dn, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(d2, d2, d1+d2, d2-d1, daxis-dsign*iny, dsign*(daxis-inx), &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(d2, d2, d1+d2, d2-d1, daxis-inx, dsign*daxis-iny+dn, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void DiagOpenPillowcaseSymm::getparams(char *s)
{ sprintf(s, "%d %d %d %d", d1, d2, dsign, daxis); }

unsigned long long DiagOpenPillowcaseSymm::getHash()
{ return (unsigned long long)getID() + 10*d1 + 100*d2 + 1000*dsign + 10000*daxis; }

int DiagOpenPillowcaseSymm::getTileX()
{ return 2*d1*d2/gcd(d1, d2); }

int DiagOpenPillowcaseSymm::getTileY()
{ return getTileX(); }

bool DiagOpenPillowcaseSymm::operator ==(const DiagOpenPillowcaseSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax && d1==s2.d1
		&& d2==s2.d2 && dsign==s2.dsign && daxis==s2.daxis;
}

//---------- (7) Orthogonal nonorientable football --------------//
OrthoFootballSymm::OrthoFootballSymm()
{ xmax=ymax=-1; }

OrthoFootballSymm::OrthoFootballSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d %d", &x, &y, &xaxis, &yaxis); }

OrthoFootballSymm OrthoFootballSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<x<=2*xmax-3; 0<y<=ymax/2.
	int x, y, xaxis, yaxis;
	
	x=1+RM.rand()%(2*xmax-3);
	y=1+RM.rand()%(ymax/2);
	xaxis=RM.rand()%2;
	yaxis=RM.rand()%2;
	
	OrthoFootballSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.x=x; s.y=y;
	s.xaxis=xaxis; s.yaxis=yaxis;
	return s;
}

void OrthoFootballSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (2*x, 0), (0, 2*y).
	//Axis of glide-reflection at x = xaxis/2.
	//Center of 180-degree rotation at ((x+xaxis)/2, (y+yaxis)/2).
	moveToBasicArea(0, 2*y, 2*x, 0, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(0, 2*y, 2*x, 0, xaxis-inx, iny+y, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*y, 2*x, 0, x+xaxis-inx, y+yaxis-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*y, 2*x, 0, inx-x, yaxis-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void OrthoFootballSymm::getparams(char *s)
{ sprintf(s, "%d %d %d %d", x, y, xaxis, yaxis); }

unsigned long long OrthoFootballSymm::getHash()
{ return (unsigned long long)getID() + 10*x + 100*y + 1000*xaxis + 10000*yaxis; }

int OrthoFootballSymm::getTileX()
{ return 2*x; }

int OrthoFootballSymm::getTileY()
{ return 2*y; }

bool OrthoFootballSymm::operator ==(const OrthoFootballSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax && x==s2.x
		&& y==s2.y && xaxis==s2.xaxis && yaxis==s2.yaxis;
}

//---------- (-7) Diagonal nonorientable football ---------------//
DiagFootballSymm::DiagFootballSymm()
{ xmax=ymax=-1; }

DiagFootballSymm::DiagFootballSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d", &d1, &d2, &daxis); }

DiagFootballSymm DiagFootballSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<d1<=d2<=xmax+1; d2<=ymax.
	int d1, d2, daxis;
	
	int m=min(xmax+1, ymax);
	d1=1+RM.rand()%m;
	d2=1+RM.rand()%m;
	if (d1>d2)
	{ int temp=d1; d1=d2; d2=temp; }
	daxis=RM.rand()%2;
	
	DiagFootballSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.d1=d1; s.d2=d2; s.daxis=daxis;
	return s;
}

void DiagFootballSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (d1, -d1), (d2, d2).
	//Axis of glide-reflection y = x + (d2 mod 2)/2.
	//Center of 180-degree rotation at (daxis/2 + (d1 + (d1 mod 2))/4 + (d2 - (d2 mod 2))/4,
	//   daxis/2 - (d1 - (d1 mod 2))/4 + (d2 + (d2 mod 2))/4).
	moveToBasicArea(d2, d2, d1+d2, d2-d1, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(d2, d2, d1+d2, d2-d1, iny - (d2+1)/2, inx - d2/2, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(d2, d2, d1+d2, d2-d1, d2/2 + (d1+1)/2 + daxis - inx,
		(d2+1)/2 - d1/2 + daxis - iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(d2, d2, d1+d2, d2-d1, daxis - d1/2 - iny, (d1+1)/2 + daxis - inx,
		&outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void DiagFootballSymm::getparams(char *s)
{ sprintf(s, "%d %d %d", d1, d2, daxis); }

unsigned long long DiagFootballSymm::getHash()
{ return (unsigned long long)getID() + 10*d1 + 100*d2 + 1000*daxis; }

int DiagFootballSymm::getTileX()
{ return 2*d1*d2/gcd(d1, d2); }

int DiagFootballSymm::getTileY()
{ return getTileX(); }

bool DiagFootballSymm::operator ==(const DiagFootballSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& d1==s2.d1 && d2==s2.d2 && daxis==s2.daxis;
}

//---------- (8) Orthogonal Mobius strip ------------------------//
OrthoMobiusSymm::OrthoMobiusSymm()
{ xmax=ymax=-1; }

OrthoMobiusSymm::OrthoMobiusSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d", &x, &y, &xaxis); }

OrthoMobiusSymm OrthoMobiusSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<x<=2*xmax-3; 0<y<=ymax/2.
	int x, y, xaxis;
	
	x=1+RM.rand()%(2*xmax-3);
	y=1+RM.rand()%(ymax/2);
	xaxis=RM.rand()%2;
	
	OrthoMobiusSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.x=x; s.y=y; s.xaxis=xaxis;
	return s;
}

void OrthoMobiusSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (x, y), (-x, y).
	//Axis of reflection at x = xaxis/2.
	moveToBasicArea(0, 2*y, x, y, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(0, 2*y, x, y, xaxis-inx, iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void OrthoMobiusSymm::getparams(char *s)
{ sprintf(s, "%d %d %d", x, y, xaxis); }

unsigned long long OrthoMobiusSymm::getHash()
{ return (unsigned long long)getID() + 10*x + 100*y + 1000*xaxis; }

int OrthoMobiusSymm::getTileX()
{ return 2*x; }

int OrthoMobiusSymm::getTileY()
{ return 2*y; }

bool OrthoMobiusSymm::operator ==(const OrthoMobiusSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax && x==s2.x
		&& y==s2.y && xaxis==s2.xaxis;
}

//---------- (-8) Diagonal Mobius strip -------------------------//
DiagMobiusSymm::DiagMobiusSymm()
{ xmax=ymax=-1; }

DiagMobiusSymm::DiagMobiusSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d", &x, &y, &sign); }

DiagMobiusSymm DiagMobiusSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints are:
	//0<=x<y<=xmax; y<=ymax.
	int x, y, sign;
	
	int m=min(xmax, ymax);
	x=RM.rand()%(m+1);
	do
	{ y=RM.rand()%(m+1); }
	while (x==y);
	if (x>y)
	{ int temp=x; x=y; y=temp; }
	sign=1-2*(RM.rand()%2); //1 or -1
	
	DiagMobiusSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.x=x; s.y=y; s.sign=sign;
	return s;
}

void DiagMobiusSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (x, y), (y, x).
	//If sign==1, axis of reflection at y = x,
	//   else, axis of reflection at x + y = 0.
	moveToBasicArea(x, y, y, x, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(x, y, y, x, sign*iny, sign*inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void DiagMobiusSymm::getparams(char *s)
{ sprintf(s, "%d %d %d", x, y, sign); }

unsigned long long DiagMobiusSymm::getHash()
{ return (unsigned long long)getID() + 10*x + 100*y + 1000*sign; }

int DiagMobiusSymm::getTileX()
{ return (y*y - x*x)/gcd(x, y); }

int DiagMobiusSymm::getTileY()
{ return getTileX(); }

bool DiagMobiusSymm::operator ==(const DiagMobiusSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& x==s2.x && y==s2.y && sign==s2.sign;
}

//---------- (9) Orthogonal turnover slit along hypotenuse ------//
OrthoTurnoverHypoSymm::OrthoTurnoverHypoSymm()
{ xmax=ymax=-1; }

OrthoTurnoverHypoSymm::OrthoTurnoverHypoSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d %d", &x, &y, &xaxis, &yaxis); }

OrthoTurnoverHypoSymm OrthoTurnoverHypoSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<x<=2*xmax-3; 0<y<=ymax-1.
	int x, y, xaxis, yaxis;
	
	x=1+RM.rand()%(2*xmax-3);
	y=1+RM.rand()%(ymax-1);
	xaxis=RM.rand()%2;
	yaxis=RM.rand()%2;
	
	OrthoTurnoverHypoSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.x=x; s.y=y;
	s.xaxis=xaxis; s.yaxis=yaxis;
	return s;
}

void OrthoTurnoverHypoSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (x, y), (-x, y).
	//Axes of reflection at x = xaxis/2, y = yaxis/2.
	moveToBasicArea(0, 2*y, x, y, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(0, 2*y, x, y, xaxis-inx, iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*y, x, y, inx, yaxis-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*y, x, y, xaxis-inx, yaxis-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void OrthoTurnoverHypoSymm::getparams(char *s)
{ sprintf(s, "%d %d %d %d", x, y, xaxis, yaxis); }

unsigned long long OrthoTurnoverHypoSymm::getHash()
{ return (unsigned long long)getID() + 10*x + 100*y + 1000*xaxis + 10000*yaxis; }

int OrthoTurnoverHypoSymm::getTileX()
{ return 2*x; }

int OrthoTurnoverHypoSymm::getTileY()
{ return 2*y; }

bool OrthoTurnoverHypoSymm::operator ==(const OrthoTurnoverHypoSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax && x==s2.x
		&& y==s2.y && xaxis==s2.xaxis && yaxis==s2.yaxis;
}

//---------- (-9) Diagonal turnover slit along hypotenuse -------//
DiagTurnoverHypoSymm::DiagTurnoverHypoSymm()
{ xmax=ymax=-1; }

DiagTurnoverHypoSymm::DiagTurnoverHypoSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d", &x, &y, &axis); }

DiagTurnoverHypoSymm DiagTurnoverHypoSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints are:
	//0<=x<y<=xmax; y<=ymax.
	int x, y, axis;
	
	int m=min(xmax, ymax);
	x=RM.rand()%(m+1);
	do
	{ y=RM.rand()%(m+1); }
	while (x==y);
	if (x>y)
	{ int temp=x; x=y; y=temp; }
	axis=RM.rand()%2;
	
	DiagTurnoverHypoSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.x=x; s.y=y; s.axis=axis;
	return s;
}

void DiagTurnoverHypoSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (x, y), (y, x).
	//Axes of reflection at y = x, x + y = axis.
	moveToBasicArea(x, y, y, x, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(x, y, y, x, iny, inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(x, y, y, x, axis-inx, axis-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(x, y, y, x, axis-iny, axis-inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void DiagTurnoverHypoSymm::getparams(char *s)
{ sprintf(s, "%d %d %d", x, y, axis); }

unsigned long long DiagTurnoverHypoSymm::getHash()
{ return (unsigned long long)getID() + 10*x + 100*y + 1000*axis; }

int DiagTurnoverHypoSymm::getTileX()
{ return (y*y - x*x)/gcd(x, y); }

int DiagTurnoverHypoSymm::getTileY()
{ return getTileX(); }

bool DiagTurnoverHypoSymm::operator ==(const DiagTurnoverHypoSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& x==s2.x && y==s2.y && axis==s2.axis;
}

//---------- (10) Turnover --------------------------------------//
TurnoverSymm::TurnoverSymm()
{ xmax=ymax=-1; }

TurnoverSymm::TurnoverSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d %d", &x, &y, &center); }

TurnoverSymm TurnoverSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<=x<=y<=xmax-1; 0<y<=ymax.
	int x, y, center;
	
	int m=min(xmax-1, ymax);
	y=1+RM.rand()%m;
	x=RM.rand()%(y+1);
	center=RM.rand()%2;
	
	TurnoverSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.x=x; s.y=y; s.center=center;
	return s;
}

void TurnoverSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (x, y), (y, -x).
	//Center of 90-degree rotation at (center/2, center/2).
	int x1=x, y1=y, x2=y, y2=-x;
	if (y2<0)
	{ x2+=x1; y2+=y1; }
	moveToBasicArea(x1, y1, x2, y2, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(x1, y1, x2, y2, iny, center-inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(x1, y1, x2, y2, center-inx, center-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(x1, y1, x2, y2, center-iny, inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void TurnoverSymm::getparams(char *s)
{ sprintf(s, "%d %d %d", x, y, center); }

unsigned long long TurnoverSymm::getHash()
{ return (unsigned long long)getID() + 10*x + 100*y + 1000*center; }

int TurnoverSymm::getTileX()
{ return (x*x + y*y)/gcd(x, y); }

int TurnoverSymm::getTileY()
{ return getTileX(); }

bool TurnoverSymm::operator ==(const TurnoverSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& x==s2.x && y==s2.y && center==s2.center;
}

//---------- (11) Orthogonal triangle ---------------------------//
OrthoTriangleSymm::OrthoTriangleSymm()
{ xmax=ymax=-1; }

OrthoTriangleSymm::OrthoTriangleSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d", &size, &center); }

OrthoTriangleSymm OrthoTriangleSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<size<=2*xmax-2; size<=2*ymax-2.
	int size, center;
	
	int m=min(2*xmax-2, 2*ymax-2);
	size=1+RM.rand()%m;
	center=RM.rand()%2;
	
	OrthoTriangleSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.size=size; s.center=center;
	return s;
}

void OrthoTriangleSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (size, 0), (0, size).
	//Center of 90-degree rotation at (center/2, center/2).
	//Axis of reflection at x = center/2.
	moveToBasicArea(0, size, size, 0, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(0, size, size, 0, iny, inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, size, size, 0, iny, center-inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, size, size, 0, inx, center-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, size, size, 0, center-inx, center-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, size, size, 0, center-iny, center-inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, size, size, 0, center-iny, inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, size, size, 0, center-inx, iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void OrthoTriangleSymm::getparams(char *s)
{ sprintf(s, "%d %d", size, center); }

unsigned long long OrthoTriangleSymm::getHash()
{ return (unsigned long long)getID() + 10*size + 100*center; }

int OrthoTriangleSymm::getTileX()
{ return size; }

int OrthoTriangleSymm::getTileY()
{ return getTileX(); }

bool OrthoTriangleSymm::operator ==(const OrthoTriangleSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& size==s2.size && center==s2.center;
}

//---------- (-11) Diagonal triangle ----------------------------//
DiagTriangleSymm::DiagTriangleSymm()
{ xmax=ymax=-1; }

DiagTriangleSymm::DiagTriangleSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d", &size, &center); }

DiagTriangleSymm DiagTriangleSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<size<=xmax-1; size<=ymax.
	int size, center;
	
	int m=min(xmax-1, ymax);
	size=1+RM.rand()%m;
	center=RM.rand()%2;
	
	DiagTriangleSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.size=size; s.center=center;
	return s;
}

void DiagTriangleSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (size, size), (-size, size).
	//Center of 90-degree rotation at (center/2, center/2).
	//Axis of reflection at x = center/2.
	moveToBasicArea(size, size, 2*size, 0, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(size, size, 2*size, 0, iny, inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(size, size, 2*size, 0, iny, center-inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(size, size, 2*size, 0, inx, center-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(size, size, 2*size, 0, center-inx, center-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(size, size, 2*size, 0, center-iny, center-inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(size, size, 2*size, 0, center-iny, inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(size, size, 2*size, 0, center-inx, iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void DiagTriangleSymm::getparams(char *s)
{ sprintf(s, "%d %d", size, center); }

unsigned long long DiagTriangleSymm::getHash()
{ return (unsigned long long)getID() + 10*size + 100*center; }

int DiagTriangleSymm::getTileX()
{ return 2*size; }

int DiagTriangleSymm::getTileY()
{ return getTileX(); }

bool DiagTriangleSymm::operator ==(const DiagTriangleSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& size==s2.size && center==s2.center;
}

//---------- (12) Orthogonal turnover slit along leg ------------//
OrthoTurnoverLegSymm::OrthoTurnoverLegSymm()
{ xmax=ymax=-1; }

OrthoTurnoverLegSymm::OrthoTurnoverLegSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d", &size, &center); }

OrthoTurnoverLegSymm OrthoTurnoverLegSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<size<=2*xmax-2; size<=(ymax+2)/2.
	int size, center;
	
	int m=min(2*xmax-2, (ymax+2)/2);
	size=1+RM.rand()%m;
	center=RM.rand()%2;
	
	OrthoTurnoverLegSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.size=size; s.center=center;
	return s;
}

void OrthoTurnoverLegSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (size, size), (-size, size).
	//Center of 90-degree rotation at (center/2, center/2).
	//Axis of reflection at x = (size+center)/2.
	moveToBasicArea(0, 2*size, size, size, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(0, 2*size, size, size, iny, center-inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*size, size, size, center-inx, center-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*size, size, size, center-iny, inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*size, size, size, size+center-inx, iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*size, size, size, iny, inx-size, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*size, size, size, inx-size, center-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*size, size, size, center-iny, size+center-inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void OrthoTurnoverLegSymm::getparams(char *s)
{ sprintf(s, "%d %d", size, center); }

unsigned long long OrthoTurnoverLegSymm::getHash()
{ return (unsigned long long)getID() + 10*size + 100*center; }

int OrthoTurnoverLegSymm::getTileX()
{ return 2*size; }

int OrthoTurnoverLegSymm::getTileY()
{ return getTileX(); }

bool OrthoTurnoverLegSymm::operator ==(const OrthoTurnoverLegSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& size==s2.size && center==s2.center;
}

//---------- (-12) Diagonal turnover slit along leg -------------//
DiagTurnoverLegSymm::DiagTurnoverLegSymm()
{ xmax=ymax=-1; }

DiagTurnoverLegSymm::DiagTurnoverLegSymm(int xm, int ym, char *params)
{ xmax=xm; ymax=ym; sscanf(params, "%d %d", &size, &center); }

DiagTurnoverLegSymm DiagTurnoverLegSymm::getRandomSymm(int xmax, int ymax)
{
	//Constraints (determined experimentally) are:
	//0<size<=2*xmax-2; size<=ymax/2.
	int size, center;
	
	int m=min(2*xmax-2, ymax/2);
	size=1+RM.rand()%m;
	center=RM.rand()%2;
	
	DiagTurnoverLegSymm s;
	s.xmax=xmax; s.ymax=ymax;
	s.size=size; s.center=center;
	return s;
}

void DiagTurnoverLegSymm::getcell(int inx, int iny, int *outx, int *outy)
{
	assert(xmax!=-1);
	//Vectors of repetition (2*size, 0), (0, 2*size).
	//Center of 90-degree rotation at (center/2, center/2).
	//Axis of reflection at x + y = size+center.
	moveToBasicArea(0, 2*size, 2*size, 0, inx, iny, outx, outy);
	int outx2, outy2;
	moveToBasicArea(0, 2*size, 2*size, 0, iny, center-inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*size, 2*size, 0, center-inx, center-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*size, 2*size, 0, center-iny, inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*size, 2*size, 0, size+center-iny, size+center-inx, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*size, 2*size, 0, size+center-inx, iny-size, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*size, 2*size, 0, iny-size, inx-size, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	moveToBasicArea(0, 2*size, 2*size, 0, inx-size, size+center-iny, &outx2, &outy2);
	pickSmaller(outx, outy, outx2, outy2);
	assert(*outx>=0 && *outx<xmax && *outy>=0 && *outy<ymax);
}

void DiagTurnoverLegSymm::getparams(char *s)
{ sprintf(s, "%d %d", size, center); }

unsigned long long DiagTurnoverLegSymm::getHash()
{ return (unsigned long long)getID() + 10*size + 100*center; }

int DiagTurnoverLegSymm::getTileX()
{ return 2*size; }

int DiagTurnoverLegSymm::getTileY()
{ return getTileX(); }

bool DiagTurnoverLegSymm::operator ==(const DiagTurnoverLegSymm &s2)
{
	return xmax==s2.xmax && ymax==s2.ymax
		&& size==s2.size && center==s2.center;
}